#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Игра "Пасьянс тысячелистник", version 0.2
# © copyleft http://www.script-coding.info/
# Автор: http://www.script-coding.info/resume.html

import sys, os, glob, stat, locale, random, datetime, sqlite3, webbrowser
from PyQt4 import QtGui, QtCore

class MainWindow(QtGui.QMainWindow):
    # self.home          # каталог скрипта (тип строка)
    # self.settings      # объект для сохранения настроек приложения (тип QSettings)
    # self.lang          # текущая языковая настройка (тип строка)
    # self.tr            # объект интернационализации (тип Translator)
    # self.pack          # текущая колода (тип строка)
    # self.scene         # объект сцены для отображения элементов-рисунков (тип Scene)
    # self.gameType      # вид текущей игры (36 или 48); если нет игры - 0
    # self.strategy_1    # стратегия текущей игры (номинал, строка от '01' до '13' или None)
    # self.strategy_2    # стратегия текущей игры (направление, число -1 или 1, или 0)
    # self.dictCardNames # названия карт во множественном числе, для строки состояния
    
    # self.lstMoves      # список сделанных ходов текущей игры, отмена ходов - по LIFO
    # элемент списка - список из 6 элементов: ссылка на Element (новая перенесённая карта),
    # старая позиция карты в игре (Element.position), старые позиции карты на сцене (Element.x, Element.y),
    # признак изменения стратегии strategy_1 (булево), признак изменения стратегии strategy_2 (булево)
    
    # self.numMoves      # количество сделанных ходов текущей игры;
    # может быть больше, чем len(lstMoves), т.к. отмены ходов считаются ходами!
    
    # self.numShuffles   # количество пересдач (перемешиваний) текущей игры
    
    def __init__(self):
        QtGui.QWidget.__init__(self)
        
        self.home = os.path.realpath(os.path.dirname(sys.argv[0])) # каталог скрипта
        self.home = unicode(self.home, locale.getpreferredencoding())
        self.home = self.home.replace('\\', '/')
        self.setWindowIcon(QtGui.QIcon(self.home + '/pics/icon.jpg')) # иконка окна
        menubar = self.menuBar() # строка меню
        
        self.gameType = 0
        self.strategy_1 = None
        self.strategy_2 = 0
        self.lstMoves = []
        self.numMoves = 0
        self.numShuffles = 0
        
        # объект для сохранения настроек приложения:
        self.settings = QtCore.QSettings('script-coding.info', 'milfoil')
        
        # восстановление языковой настройки:
        self.lang = self.settings.value('lang', QtCore.QVariant(u'milfoil_en_EN.qm')).toString()
        self.tr = Translator(None, self.lang)
        # восстановление настройки выбора колоды:
        self.pack = self.settings.value('pack', QtCore.QVariant(u'Standard')).toString()
        
        # заголовок окна:
        self.setWindowTitle(self.tr.translate('Milfoil patience'))
        # восстановление настройки геометрии окна:
        self.setGeometry(self.settings.value('geometry', QtCore.QVariant(QtCore.QRect(30, 30, 670, 670))).toRect())
        
        # создание сцены для отображения элементов-рисунков:
        self.scene = Scene()
        # создание виджета представления для отображения сцены:
        self.view = QtGui.QGraphicsView(self.scene, self)
        # параметры качества прорисовки для виджета представления:
        self.view.setRenderHints(QtGui.QPainter.Antialiasing | QtGui.QPainter.SmoothPixmapTransform)
        self.view.setBackgroundBrush(QtGui.QColor(0, 128, 64)) # цвет фона представления
        self.setCentralWidget(self.view) # размещение виджета представления в главном окне
        
        # названия карт, для строки состояния:
        self.dictCardNames = {'01':'aces', '02':'deuces', '03':'threes', '04':'fours', '05':'fives',
                              '06':'sixes', '07':'sevens', '08':'eightes', '09':'nines',
                              '10':'tens', '11':'jacks', '12':'queens', '13':'kings'}
                              
        self.statusBar = self.statusBar() # создание строки состояния
        self.status = QtGui.QLabel(self.statusBar)
        self.statusBar.addWidget(self.status, 1)
        self.setStatusText()
        
        # меню "Language choice":
        menuLang = menubar.addMenu(self.tr.translate('Language choice'))
        
        pointName = self.tr.translate('By default')
        point = QtGui.QAction(pointName, self)
        self.connect(point, QtCore.SIGNAL('triggered()'), self.lang_menu)
        point.setDisabled(self.lang == u'milfoil_en_EN.qm')
        menuLang.addAction(point)
        
        for filename in glob.glob(self.home + '/langs/milfoil_*_*.qm'):
            pointName = os.path.basename(filename)
            point = QtGui.QAction(pointName, self)
            self.connect(point, QtCore.SIGNAL('triggered()'), self.lang_menu)
            point.setDisabled(point.text() == self.lang)
            menuLang.addAction(point)
        
        # меню "Pack choice":
        menuPack = menubar.addMenu(self.tr.translate('Pack choice'))
        for filename in os.listdir(self.home + '/pics/packs'):
            if stat.S_ISDIR(os.stat(self.home + '/pics/packs/' + filename).st_mode): # это папка
                #loc = locale.getlocale()[1] if locale.getlocale()[1] else 'cp1251'
                #point = QtGui.QAction(unicode(filename, loc), self)
                point = QtGui.QAction(filename, self)
                self.connect(point, QtCore.SIGNAL('triggered()'), self.pack_menu)
                point.setDisabled(point.text() == self.pack)
                menuPack.addAction(point)
                
        # меню "Game"
        menuGame = menubar.addMenu(self.tr.translate('Game'))
        
        pointName = self.tr.translate('New game (36 cards)')
        point = QtGui.QAction(pointName, self)
        point.setShortcut('Ctrl+1')
        self.connect(point, QtCore.SIGNAL('triggered()'), self.new_game_36)
        menuGame.addAction(point)
        
        pointName = self.tr.translate('New game (48 cards)')
        point = QtGui.QAction(pointName, self)
        point.setShortcut('Ctrl+2')
        self.connect(point, QtCore.SIGNAL('triggered()'), self.new_game_48)
        menuGame.addAction(point)
        
        menuGame.addSeparator()
        
        pointName = self.tr.translate('Shuffle the cards')
        point = QtGui.QAction(pointName, self)
        point.setShortcut('Ctrl+N')
        point.setDisabled(True)
        self.connect(point, QtCore.SIGNAL('triggered()'), self.shuffleTheCards)
        menuGame.addAction(point)
        
        pointName = self.tr.translate('Undo the last move')
        point = QtGui.QAction(pointName, self)
        point.setShortcut('Backspace')
        point.setDisabled(True)
        self.connect(point, QtCore.SIGNAL('triggered()'), self.undoTheLastMove)
        menuGame.addAction(point)
        
        menuGame.addSeparator()
        
        pointName = self.tr.translate('High scores')
        point = QtGui.QAction(pointName, self)
        self.connect(point, QtCore.SIGNAL('triggered()'), self.highScores)
        menuGame.addAction(point)
        
        # меню "Help"
        menuHelp = menubar.addMenu(self.tr.translate('Help'))
        
        pointName = self.tr.translate('Guidelines')
        point = QtGui.QAction(pointName, self)
        point.setShortcut('F1')
        self.connect(point, QtCore.SIGNAL('triggered()'), self.Guidelines)
        menuHelp.addAction(point)
        
    def Guidelines(self):
        try:
            webbrowser.open(self.home + '/langs/' + self.lang[:-2].toUtf8() + 'html')
        except:
            caption = self.tr.translate('Milfoil patience')
            text = self.tr.translate('There isn\'t Help for selected language.')
            QtGui.QMessageBox.information(self, caption, text, QtGui.QMessageBox.Ok)
        
    def gameOver(self):
        # Процедура завершения игры.
        game = str(self.gameType)
        self.gameType = 0
        self.shuffleAvailability(False) # деактивизация меню "Shuffle the cards"
        self.lstMoves = [] # деактивизация меню "Undo the last move"
        # надпись на сцене:
        font = QtGui.QFont()
        font.setPixelSize(18)
        font.setBold(True)
        self.scene.addText(self.tr.translate('The patience is completed!'), font)
        # диалог для получения имени:
        dlg = Dialog(self)
        if dlg.exec_(): # нажата кнопка "Ок"
            nm = unicode(dlg.ln_edit1.text())
            if len(nm) == 0:
                nm = unicode(self.tr.translate('Anonymous'))
            # таблица рекордов:
            # nm - имя пользователя
            # dt - дата
            # game - вид игры (36 или 48)
            # shuffles - количество пересдач
            # moves - количество ходов
            conn = sqlite3.connect(self.home + '/data/scores.dat')
            # создаём таблицу, если её нет:
            conn.execute('create table if not exists scores(nm, dt, game, shuffles, moves)')
            dt = datetime.datetime.now().strftime('%Y-%m-%d %p %I:%M:%S')
            shuffles = str(self.numShuffles)
            moves = str(self.numMoves)
            # данные для записи в базу данных:
            data = [(nm, dt, game, shuffles, moves)]
            # запись в базу данных:
            conn.executemany('insert into scores(nm, dt, game, shuffles, moves) values(?, ?, ?, ?, ?)', data)
            conn.commit() # сохранение изменений в базе данных
            conn.close() # закрытие базы данных
        dlg.destroy()
        self.highScores()
        
    def highScores(self):
        # Обработчик пункта меню "High scores".
        try:
            conn = sqlite3.connect(self.home + '/data/scores.dat')
            for row in conn.execute("select count(*) from scores"):
                r = row[0]
                break
            tbl = QtGui.QTableWidget(r, 5) # r строк, 5 столбцов
            lst = [self.tr.translate('Game'), self.tr.translate('Name'),
                   self.tr.translate('Shuffles'), self.tr.translate('Moves'),
                   self.tr.translate('Time')]
            lst = QtCore.QStringList(lst)
            tbl.setHorizontalHeaderLabels(lst)
            i = 0
            for row in conn.execute("select * from scores order by game desc, shuffles, moves, dt, nm"):
                j = 0
                for fld in (2, 0, 3, 4, 1): # игра, имя, пересдачи, ходы, время
                    item = QtGui.QTableWidgetItem(QtCore.QString(row[fld]))
                    item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                    tbl.setItem(i, j, item)
                    j += 1
                i += 1
            conn.close()  # закрытие базы данных
            tbl.setSortingEnabled(True)
            dlg = QtGui.QDialog(self)
            dlg.resize(550, 250)
            dlg.setMinimumWidth(300) # минимальная ширина окна диалога
            dlg.setMinimumHeight(150) # минимальная высота окна диалога
            dlg.setWindowTitle(self.tr.translate('High scores'))
            boxlay = QtGui.QHBoxLayout(dlg)
            boxlay.addWidget(tbl)
            dlg.exec_()
            dlg.destroy()
        except:
            caption = self.tr.translate('Milfoil patience')
            text = self.tr.translate('No scores.')
            QtGui.QMessageBox.information(self, caption, text, QtGui.QMessageBox.Ok)
        
    def undoTheLastMove(self):
        # Обработчик пункта меню "Undo the last move".
        move = self.lstMoves.pop() # выталкиваем последний ход
        card = move[0] # перемещённая карта
        position = move[1] # старая позиция карты в игре
        x = move[2] # старые позиции карты на сцене
        y = move[3]
        # восстановление стратегии
        if move[4]:
            self.strategy_1 = None
        if move[5]:
            self.strategy_2 = 0
        
        # карта на новом месте
        newCard = self.createCardItem(card.nominal)
        newCard.setPos(x, y)
        if position in [3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 36, 39, 42, 45, 48]:
            zVal = 25
        elif position in [1, 4, 7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 40, 41, 46]:
            zVal = -25
            newCard.dropAccept = True
        else:
            zVal = 0
            newCard.dropAccept = True
        newCard.rotate(zVal)
        newCard.setZValue(zVal)
        newCard.position = position
        newCard.x = x
        newCard.y = y
        newCard.cursorType = QtCore.Qt.OpenHandCursor # карту можно перетаскивать
        newCard.setCursor(QtCore.Qt.OpenHandCursor)
        self.scene.addItem(newCard) # добавление карты на новом месте на сцену
        
        # карта, которая лежала под ранее перемещённой
        if zVal >= 0:
            for elem in self.scene.items():
                if elem.position == position - 1:
                    elem.dropAccept = False
                    elem.cursorType = QtCore.Qt.ForbiddenCursor
                    elem.setCursor(QtCore.Qt.ForbiddenCursor)
                    break
        
        self.scene.removeItem(card) # удаление ранее перемещённой карты
        # карта, которая лежит под ранее перемещённой
        if card.position < 50:
            for elem in self.scene.items():
                if elem.position == card.position - 1:
                    elem.dropAccept = True
                    elem.cursorType = QtCore.Qt.OpenHandCursor
                    elem.setCursor(QtCore.Qt.OpenHandCursor)
                    break
        else: # ход был "на посадочное место"
            for elem in self.scene.items():
                if elem.position == card.position:
                    if card.zValue() - elem.zValue() == 1:
                        elem.dropAccept = True
                        break
                        
        self.undoTheLastMoveAvailability()
        self.numMoves += 1
        self.setStatusText()
        
    def shuffleAvailability(self, yes):
        # Процедура делает доступным или недоступным меню "Shuffle the cards",
        # в зависимости от параметра "yes" (булево).
        for act in self.menuBar().actions():
            if act.text() == self.tr.translate('Game'):
                for point in act.menu().actions():
                    if point.text() == self.tr.translate('Shuffle the cards'):
                        point.setEnabled(yes)
                        break
                break
                
    def undoTheLastMoveAvailability(self):
        # Процедура делает доступным или недоступным меню "Undo the last move",
        # в зависимости от наличия сделанных ходов.
        for act in self.menuBar().actions():
            if act.text() == self.tr.translate('Game'):
                for point in act.menu().actions():
                    if point.text() == self.tr.translate('Undo the last move'):
                        if len(self.lstMoves) > 0: # есть сделанные ходы
                            point.setEnabled(True)
                        else:
                            point.setEnabled(False)
                        break
                break
        
    def setStatusText(self):
        # Процедура устанавливает текст строки состояния.
        if self.gameType == 0:
            self.status.setText(self.tr.translate('no game'))
        else:
            if self.strategy_1 == None:
                st = self.tr.translate('no strategy')
            else:
                st = self.tr.translate('strategy') + ': '
                st += self.tr.translate(self.dictCardNames[self.strategy_1])
                if self.strategy_2 != 0:
                    if self.strategy_2 < 0:
                        st += ', ' + self.tr.translate('on reduction')
                    else: # self.strategy_2 > 0
                        st += ', ' + self.tr.translate('on increase')
            # пересдачи:
            st += '; ' + self.tr.translate('shuffles') + ': ' + str(self.numShuffles)
            # ходы:
            st += '; ' + self.tr.translate('moves') + ': ' + str(self.numMoves)
            self.status.setText(st)
            
    def getList_48(self):
        # Процедура возвращает список из строковых обозначений карт, колода 48 карт.
        res = []
        for x in xrange(2, 14): # номиналы с двойки (2) по короля (13)
            for y in xrange(1, 5): # масти с 1 по 4
                res.append(('0' + str(x) + str(y))[-3:])
        return res
        
    def getList_36(self):
        # Процедура возвращает список из строковых обозначений карт, колода 36 карт.
        res = []
        for x in xrange(6, 14): # номиналы с шестёрки (6) по короля (13)
            for y in xrange(1, 5): # масти с 1 по 4
                res.append(('0' + str(x) + str(y))[-3:])
        for y in xrange(1, 5): # масти с 1 по 4 (добавляем тузов)
            res.append('01' + str(y))
        return res
        
    def createCardItem(self, designation):
        # Процедура генерирует и возвращает графический элемент карты.
        picPath = self.home + '/pics/packs/' + self.pack + '/' + designation + '.gif'
        card = Element(QtGui.QPixmap(picPath), None)
        h = card.boundingRect().height()
        w = card.boundingRect().width()
        card.setMatrix(card.matrix().scale(123.5/h, 85.0/w))
        card.nominal = designation
        return card
        
    def createPlaceItem(self):
        # Процедура генерирует и возвращает графический элемент пустого "посадочного" места для карт.
        picPath = self.home + '/pics/place.jpg'
        card = Element(QtGui.QPixmap(picPath), None)
        h = card.boundingRect().height()
        w = card.boundingRect().width()
        card.setMatrix(card.matrix().scale(123.5/h, 85.0/w))
        return card
        
    def new_game(self, lst, rowNum, grNum, width):
        # Процедура сдаёт карты для новой игры.
        # lst - список из строковых обозначений карт
        # rowNum - количество горизонтальных рядов карт
        # grNum - количество групп карт в каждом ряду
        # width - необходимая штрина окна
        self.scene.clear() # очистка сцены
        x = 0; y = 0 # позиция первой карты
        counter = 1 # счётчик сдачи
        for k in xrange(rowNum): # ряды карт
            for i in xrange(grNum): # группы карт (по 3 штуки)
                if len(lst) == 0: break
                for j in [-25, 0, 25]:
                    random.shuffle(lst) # тасуем колоду
                    card = self.createCardItem(lst.pop())
                    card.rotate(j)
                    card.setZValue(j)
                    xPos = x
                    if j == 0: yPos = y-15
                    elif j > 0: yPos = y-10
                    else: yPos = y
                    card.setPos(xPos, yPos)
                    card.x = xPos; card.y = yPos
                    card.position = counter
                    if j > 0: # верхняя карта в группе
                        card.setCursor(QtCore.Qt.OpenHandCursor)
                        card.cursorType = QtCore.Qt.OpenHandCursor
                    else:
                        card.setCursor(QtCore.Qt.ForbiddenCursor)
                        card.cursorType = QtCore.Qt.ForbiddenCursor
                    self.scene.addItem(card)
                    counter += 1
                    x += 51
            y += 144; x = 0
        if rowNum == 4: x = 195 # 48 карт
        else: x = 51 # 36 карт
        counter = 51
        for k in xrange(4): # пустые "посадочные" места для карт
            card = self.createPlaceItem()
            card.setZValue(-24)
            card.setPos(x, 433)
            card.x = x; card.y = 433
            card.position = counter
            card.dropAccept = True
            self.scene.addItem(card)
            counter += 1
            x += 120
        # размеры окна:
        if self.geometry().height() < 670: h = 670
        else: h = self.geometry().height()
        if self.geometry().width() < width: w = width
        else: w = self.geometry().width()
        self.setGeometry(self.geometry().left(), self.geometry().top(), w, h)
        # стратегия и счётчики ходов:
        self.strategy_1 = None
        self.strategy_2 = 0
        self.lstMoves = []
        self.numMoves = 0
        self.numShuffles = 0
        self.setStatusText()
        self.shuffleAvailability(True) # активизация меню "Shuffle the cards"
        
    def shuffleTheCards(self):
        # Обработчик пункта меню "Shuffle the cards".
        lst = []
        for card in self.scene.items():
            if card.position < 50: # карта на столе
                lst.append(card.nominal)
                self.scene.removeItem(card)
        if self.gameType == 36:
            rowNum = 3 # количество горизонтальных рядов карт
            grNum = 4 # количество групп карт в каждом ряду
        elif self.gameType == 48:
            rowNum = 4 # количество горизонтальных рядов карт
            grNum = 5 # количество групп карт в каждом ряду
        x = 0; y = 0 # позиция первой карты
        counter = 1 # счётчик сдачи
        for k in xrange(rowNum): # ряды карт
            for i in xrange(grNum): # группы карт (по 3 штуки)
                for j in [-25, 0, 25]:
                    if len(lst) == 0: break
                    random.shuffle(lst) # тасуем колоду
                    card = self.createCardItem(lst.pop())
                    card.rotate(j)
                    card.setZValue(j)
                    xPos = x
                    if j == 0: yPos = y-15
                    elif j > 0: yPos = y-10
                    else: yPos = y
                    card.setPos(xPos, yPos)
                    card.x = xPos; card.y = yPos
                    card.position = counter
                    if j > 0 or len(lst) == 0: # верхняя карта в группе
                        card.setCursor(QtCore.Qt.OpenHandCursor)
                        card.cursorType = QtCore.Qt.OpenHandCursor
                    else:
                        card.setCursor(QtCore.Qt.ForbiddenCursor)
                        card.cursorType = QtCore.Qt.ForbiddenCursor
                    if len(lst) == 0 and j <= 0: # верхняя карта в последней, незавершённой группе
                        card.dropAccept = True
                    self.scene.addItem(card)
                    counter += 1
                    x += 51
            y += 144; x = 0
        self.numShuffles += 1 # счётчик пересдач
        self.setStatusText()
        self.lstMoves = [] # деактивизация меню "Undo the last move"
        self.undoTheLastMoveAvailability()
        
    def new_game_48(self):
        # Обработчик пункта меню "New game (48 cards)".
        self.gameType = 48
        self.new_game(self.getList_48(), 4, 5, 830)
        
    def new_game_36(self):
        # Обработчик пункта меню "New game (36 cards)".
        self.gameType = 36
        self.new_game(self.getList_36(), 3, 4, 670)
        
    def lang_menu(self):
        # Обработчик любого пункта колонки меню "Language choice".
        pointName = self.sender().text()
        if pointName == self.tr.translate('By default'):
            pointName = u'milfoil_en_EN.qm'
        self.lang = pointName
        
        name = self.tr.translate('Language choice')
        for act in self.menuBar().actions():
            if act.text() == name:
                for point in act.menu().actions():
                    point.setDisabled(point == self.sender())
                break
                
        caption = self.tr.translate('Milfoil patience')
        text = self.tr.translate('In order for language setting to come into force, restart the application.')
        QtGui.QMessageBox.information(self, caption, text, QtGui.QMessageBox.Ok)
        
    def pack_menu(self):
        # Обработчик любого пункта колонки меню "Pack choice".
        self.pack = self.sender().text()
        
        name = self.tr.translate('Pack choice')
        for act in self.menuBar().actions():
            if act.text() == name:
                for point in act.menu().actions():
                    point.setDisabled(point == self.sender())
                break
                
    def closeEvent(self, event):
        # сохранение языковой настройки:
        self.settings.setValue('lang', QtCore.QVariant(self.lang))
        # сохранение настройки выбора колоды:
        self.settings.setValue('pack', QtCore.QVariant(self.pack))
        # сохранение настройки геометрии окна:
        self.settings.setValue('geometry', QtCore.QVariant(self.geometry()))
        
class Translator(QtCore.QTranslator):
    def __init__(self, parent=None, lang='milfoil_en_EN.qm'):
        QtCore.QTranslator.__init__(self, parent)
        home = os.path.realpath(os.path.dirname(sys.argv[0])) # каталог скрипта
        home = unicode(home, locale.getpreferredencoding())
        self.load(home + '/langs/' + lang)
        
    def translate(self, sourceText):
        res = QtCore.QTranslator.translate(self, 'milfoil', sourceText)
        if len(res) == 0:
            res = QtCore.QString(sourceText)
        return res
        
class Scene(QtGui.QGraphicsScene):
    def __init__(self, parent = None):
        QtGui.QGraphicsScene.__init__(self, parent)
        self.dragOverElement = None # элемент, над которым в данный момент осуществляется drag-and-drop
        
    def dragEnterEvent(self, event):
        # Операция drag-and-drop входит в область сцены.
        try:
            card = event.mimeData().Element # перетаскиваемая карта
            # временный "затемнённый" рисунок перетаскиваемой картинки:
            if card.tempPixmap == None:
                card.tempPixmap = card.pixmap()
                pix = card.pixmap()
                painter = QtGui.QPainter()
                painter.begin(pix)
                painter.fillRect(pix.rect(), QtGui.QColor(127, 127, 127, 127))
                painter.end()
                card.setPixmap(pix)
        except: # на сцену перетащили что-то из-за пределов приложения
            pass
        
    def dragLeaveEvent(self, event):
        # Операция drag-and-drop покидает область сцены.
        try:
            card = event.mimeData().Element # перетаскиваемая карта
            if card.tempPixmap != None:
                card.setPixmap(card.tempPixmap)
                card.tempPixmap = None
        except: # на сцену перетащили что-то из-за пределов приложения
            pass
            
    def dragMoveEvent(self, event):
        # В процессе операции drag-and-drop.
        try:
            # элемент, над которым в данный момент осуществляется drag-and-drop:
            card = self.itemAt(event.scenePos())
            if card != self.dragOverElement:
                if card == None and self.dragOverElement != None:
                    self.dragOverElement.dragLeaveEvent(event)
                elif card != None and self.dragOverElement == None:
                    card.dragEnterEvent(event)
                else:
                    self.dragOverElement.dragLeaveEvent(event)
                    card.dragEnterEvent(event)
                self.dragOverElement = card
        except: # на сцену перетащили что-то из-за пределов приложения
            pass
        
    def dropEvent(self, event):
        # Завершение операции drag and drop.
        try:
            card = self.itemAt(event.scenePos())
            if card == None:
                elem = event.mimeData().Element # переносимый элемент
                elem.setPixmap(elem.tempPixmap)
                elem.tempPixmap = None
            else:
                card.dropEvent(event)
        except: # на сцену перетащили что-то из-за пределов приложения
            pass
        
class Element(QtGui.QGraphicsPixmapItem):
    def __init__(self, pixmap, parent = None, scene = None):
        QtGui.QGraphicsPixmapItem.__init__(self, pixmap, parent, scene)
        self.setTransformationMode(QtCore.Qt.SmoothTransformation) # качество прорисовки
        self.setCursor(QtCore.Qt.ForbiddenCursor) # вид курсора мыши над элементом
        self.setAcceptDrops(True) # разрешить перетаскивание
        
        # self.zValue() у нижних карт всегда -25
        # self.zValue() у средних карт всегда 0
        # self.zValue() у верхних карт всегда 25
        # self.zValue() у посадочных мест всегда -24
        # self.zValue() у карт на посадочных местах всегда от -23 до -11
        # таким образом, угол наклона карты и её позиция в группе карт однозначно определяются по её zValue()
        
        self.nominal = '-' # достоинство и масть карты
        self.position = 0 # позиция карты в игре
        # от 1 до 48 - карта на столе
        # от 51 до 54 - карта на посадочном месте
        self.x = 0; self.y = 0 # позиция карты на сцене
        # вид курсора мыши над элементом (определяет, можно ли начинать перетаскивать эту карту):
        self.cursorType = QtCore.Qt.ForbiddenCursor
        self.dropAccept = False # определяет, можно ли что-либо перетаскивать НА эту карту в принципе
        self.tempPixmap = None # временное хранилище исходной картинки карты, когда картинка модифицируется
        
    def dropEnabled(self, cardUnder, cardOver):
        # Процедура определяет, можно ли положить cardOver на cardUnder и возвращает bool.
        mainWnd = self.scene().views()[0].parent()
        if cardUnder.position < 50: # карта на столе
            if cardUnder.nominal[:2] == cardOver.nominal[:2]: # одинаковое достоинство
                return True
        else: # карта на посадочном месте
            if cardUnder.nominal == '-': # пустое посадочное место
                if mainWnd.strategy_1 == None:
                    return True
                elif cardOver.nominal[:2] == mainWnd.strategy_1:
                    return True
            elif cardUnder.nominal[2] == cardOver.nominal[2]: # одинаковая масть
                valUnder = int(cardUnder.nominal[:2])
                valOver = int(cardOver.nominal[:2])
                if abs(valUnder - valOver) == 1: # достоинство отличается на единицу
                    if mainWnd.strategy_2 == 0:
                        return True
                    elif valOver - valUnder == mainWnd.strategy_2:
                        return True
                else:
                    gameType = mainWnd.gameType # вид игры
                    if gameType == 36: # игра 36 карт
                        if mainWnd.strategy_2 == 0:
                            # туз на короля или наоборот, шестёрка на туза или наоборот:
                            if (valUnder == 13 and valOver ==  1) or \
                               (valUnder ==  1 and valOver == 13) or \
                               (valUnder ==  1 and valOver ==  6) or \
                               (valUnder ==  6 and valOver ==  1):
                                return True
                        elif mainWnd.strategy_2 == 1:
                            # туз на короля или шестёрка на туза:
                            if (valUnder == 13 and valOver ==  1) or \
                               (valUnder ==  1 and valOver ==  6):
                                return True
                        elif mainWnd.strategy_2 == -1:
                            # король на туза или туз на шестёрку:
                            if (valUnder ==  1 and valOver == 13) or \
                               (valUnder ==  6 and valOver ==  1):
                                return True
                    elif gameType == 48: # игра 48 карт
                        if mainWnd.strategy_2 == 0:
                            # двойка на короля или наоборот:
                            if (valUnder == 13 and valOver ==  2) or \
                               (valUnder ==  2 and valOver == 13):
                                return True
                        elif mainWnd.strategy_2 == 1:
                            # двойка на короля:
                            if (valUnder == 13 and valOver ==  2):
                                return True
                        elif mainWnd.strategy_2 == -1:
                            # король на двойку:
                            if (valUnder ==  2 and valOver == 13):
                                return True
        return False
        
    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton: # левая клавиша мыши
            if self.cursorType == QtCore.Qt.ForbiddenCursor: # карту нельзя перетаскивать
                event.ignore()
                return
            drag = QtGui.QDrag(event.widget()) # объект Drag
            mime = QtCore.QMimeData()
            mime.setImageData(QtCore.QVariant(self.pixmap())) # запоминаем рисунок
            mime.Element = self # запоминаем элемент, который переносится
            drag.setMimeData(mime)
            
            pix = self.pixmap()
            pix = pix.transformed(QtGui.QMatrix().scale(123.5/pix.height(), 85.0/pix.width()))
            drag.setPixmap(pix) # рисунок, отображающийся в процессе переноса
            #drag.setHotSpot(QtCore.QPoint(50.0, 72.5)) # позиция "ухватки", всегда за середину карты
            drag.setHotSpot(QtCore.QPoint(0, 0)) # позиция "ухватки", всегда левый верхний угол карты
            
            drag.start() # запуск (начало) перетаскивания
        elif event.button() == QtCore.Qt.RightButton: # правая клавиша мыши
            # Показываем все карты с таким же достоинством.
            for card in self.scene().items():
                if card.nominal[:2] == self.nominal[:2]:
                    card.tempPixmap = card.pixmap()
                    pix = card.pixmap()
                    painter = QtGui.QPainter()
                    painter.begin(pix)
                    painter.fillRect(pix.rect(), QtGui.QColor(127, 127, 127, 127))
                    painter.end()
                    card.setPixmap(pix)
                    
    def mouseDoubleClickEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton: # левая клавиша мыши
            # Пытаемся бросить карту на первое подходящее посадочное место.
            if self.cursorType == QtCore.Qt.ForbiddenCursor: # карту нельзя перетаскивать
                event.ignore()
                return
            scene = self.scene()
            mainWnd = scene.views()[0].parent()
            for card in self.scene().items():
                if card.position > 50 and card.dropAccept: # посадочное место или карта на нём
                    if self.dropEnabled(card, self):
                        newCard = mainWnd.createCardItem(self.nominal)
                        newCard.setPos(card.pos()) # установка положения нового элемента, точно над текущим элементом
                        newCard.setZValue(card.zValue() + 1) # кладём сверху
                        newCard.position = card.position # позиция карты в игре
                        newCard.x = card.x; newCard.y = card.y # позиция карты на сцене
                        newCard.cursorType = QtCore.Qt.ForbiddenCursor # карту нельзя перетаскивать
                        newCard.setCursor(QtCore.Qt.ForbiddenCursor)
                        newCard.dropAccept = True # на эту карту можно перетаскивать
                        # добавление новой карты на сцену:
                        scene.addItem(newCard)
                        # вскрытая в результате переноса карта:
                        if self.zValue() >= 0: # перенесённая карта - средняя или верхняя, иначе вскрытой карты нет
                            for uncoveredCard in scene.items():
                                if uncoveredCard.position == self.position - 1:
                                    uncoveredCard.cursorType = QtCore.Qt.OpenHandCursor # карту можно перетаскивать
                                    uncoveredCard.setCursor(QtCore.Qt.OpenHandCursor)
                                    uncoveredCard.dropAccept = True # на эту карту можно перетаскивать
                                    break
                        # удаление перенесённого элемента:
                        scene.removeItem(self)
                        # на эту карту теперь нельзя перетаскивать:
                        card.dropAccept = False
                        # эту карту теперь нельзя перетаскивать:
                        card.cursorType = QtCore.Qt.ForbiddenCursor
                        card.setCursor(QtCore.Qt.ForbiddenCursor)
                        # определение стратегии:
                        str1 = False; str2 = False # флаги изменения
                        if mainWnd.strategy_1 == None or mainWnd.strategy_2 == 0:
                            if mainWnd.strategy_1 == None:
                                mainWnd.strategy_1 = self.nominal[:2]
                                str1 = True
                            if card.nominal != '-':
                                if mainWnd.strategy_2 == 0:
                                    str2 = True
                                    diff = int(self.nominal[:2]) - int(card.nominal[:2])
                                    under = card.nominal[:2]
                                    over = self.nominal[:2]
                                    if mainWnd.gameType == 36: # игра 36 карт
                                        if diff == 1 or (under == '13' and over == '01') or (under == '01' and over == '06'):
                                            mainWnd.strategy_2 = 1
                                        else:
                                            mainWnd.strategy_2 = -1
                                    elif mainWnd.gameType == 48: # игра 48 карт
                                        if diff == 1 or (under == '13' and over == '02'):
                                            mainWnd.strategy_2 = 1
                                        else:
                                            mainWnd.strategy_2 = -1
                        # фиксация хода:
                        mainWnd.numMoves += 1 # счётчик ходов
                        mainWnd.setStatusText()
                        lstMove = [newCard, self.position, self.x, self.y, str1, str2]
                        mainWnd.lstMoves.append(lstMove)
                        # проверка завершения игры:
                        if newCard.isGameOver():
                            mainWnd.gameOver()
                        mainWnd.undoTheLastMoveAvailability()
                        break
                        
    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.RightButton: # правая клавиша мыши
            for card in self.scene().items():
                if card.nominal[:2] == self.nominal[:2]:
                    if card.tempPixmap != None:
                        card.setPixmap(card.tempPixmap)
                        card.tempPixmap = None
            
    def dragEnterEvent(self, event):
        # Операция drag-and-drop входит в область элемента.
        try:
            card = event.mimeData().Element # пертаскиваемая карта
            if self.dropAccept:
                if self.dropEnabled(self, card):
                    # временный "затемнённый" рисунок картинки:
                    if self.tempPixmap == None:
                        self.tempPixmap = self.pixmap()
                        pix = self.pixmap()
                        painter = QtGui.QPainter()
                        painter.begin(pix)
                        painter.fillRect(pix.rect(), QtGui.QColor(127, 127, 127, 127))
                        painter.end()
                        self.setPixmap(pix)
        except: # на сцену перетащили что-то из-за пределов приложения
            pass
            
    def dragLeaveEvent(self, event):
        # Операция drag-and-drop покидает область элемента.
        elem = event.mimeData().Element # переносимый элемент
        if self.tempPixmap != None and elem != self:
            self.setPixmap(self.tempPixmap)
            self.tempPixmap = None
        
    def isGameOver(self):
        # Процедура определяет, не завершена ли игра, и возвращает bool.
        for card in self.scene().items():
            if card.position < 50: # карта на столе
                return False
        return True
        
    def dropEvent(self, event):
        # Завершение операции drag and drop.
        scene = self.scene()
        mainWnd = scene.views()[0].parent()
        elem = event.mimeData().Element # перенесённый элемент
        if self.dropAccept and self != elem and self.dropEnabled(self, elem):
            # создание копии перенесённого элемента на новом месте:
            card = mainWnd.createCardItem(elem.nominal)
            if 51 <= self.position <= 54: # посадочное место или карта на нём
                card.setPos(self.pos()) # установка положения нового элемента, точно над текущим элементом
                card.setZValue(self.zValue() + 1) # кладём сверху
                card.position = self.position # позиция карты в игре
                card.x = self.x; card.y = self.y # позиция карты на сцене
                card.cursorType = QtCore.Qt.ForbiddenCursor # карту нельзя перетаскивать
                card.setCursor(QtCore.Qt.ForbiddenCursor)
                card.dropAccept = True # на эту карту можно перетаскивать
            else:
                if self.zValue() == 0: # средняя карта в группе
                    card.setPos(self.x + 51, self.y + 10) # установка положения нового элемента
                    card.x = self.x + 51; card.y = self.y + 10 # позиция карты на сцене
                    card.dropAccept = False # на эту карту нельзя перетаскивать
                    card.rotate(25)
                else: # нижняя карта в группе
                    card.setPos(self.x + 51, self.y - 15) # установка положения нового элемента
                    card.x = self.x + 51; card.y = self.y - 15 # позиция карты на сцене
                    card.dropAccept = True # на эту карту можно перетаскивать
                card.setZValue(self.zValue() + 25)
                card.position = self.position + 1
                card.cursorType = QtCore.Qt.OpenHandCursor # карту можно перетаскивать
                card.setCursor(QtCore.Qt.OpenHandCursor)
            # добавление новой карты на сцену:
            scene.addItem(card)
            # вскрытая в результате переноса карта:
            if elem.zValue() >= 0: # перенесённая карта - средняя или верхняя, иначе вскрытой карты нет
                for uncoveredCard in scene.items():
                    if uncoveredCard.position == elem.position - 1:
                        uncoveredCard.cursorType = QtCore.Qt.OpenHandCursor # карту можно перетаскивать
                        uncoveredCard.setCursor(QtCore.Qt.OpenHandCursor)
                        uncoveredCard.dropAccept = True # на эту карту можно перетаскивать
                        break
            # удаление перенесённого элемента:
            scene.removeItem(elem)
            # на эту карту теперь нельзя перетаскивать:
            self.dropAccept = False
            # эту карту теперь нельзя перетаскивать:
            self.cursorType = QtCore.Qt.ForbiddenCursor
            self.setCursor(QtCore.Qt.ForbiddenCursor)
            # восстановление рисунка:
            if self.tempPixmap != None:
                self.setPixmap(self.tempPixmap)
                self.tempPixmap = None
            # определение стратегии:
            str1 = False; str2 = False # флаги изменения
            if self.position > 50: # посадочное место или карта на нём
                if mainWnd.strategy_1 == None or mainWnd.strategy_2 == 0:
                    if mainWnd.strategy_1 == None:
                        mainWnd.strategy_1 = card.nominal[:2]
                        str1 = True
                    if mainWnd.strategy_2 == 0:
                        if self.nominal != '-':
                            str2 = True
                            diff = int(card.nominal[:2]) - int(self.nominal[:2])
                            under = self.nominal[:2]
                            over = card.nominal[:2]
                            if mainWnd.gameType == 36: # игра 36 карт
                                if diff == 1 or (under == '13' and over == '01') or (under == '01' and over == '06'):
                                    mainWnd.strategy_2 = 1
                                else:
                                    mainWnd.strategy_2 = -1
                            elif mainWnd.gameType == 48: # игра 48 карт
                                if diff == 1 or (under == '13' and over == '02'):
                                    mainWnd.strategy_2 = 1
                                else:
                                    mainWnd.strategy_2 = -1
            # фиксация хода:
            mainWnd.numMoves += 1 # счётчик ходов
            mainWnd.setStatusText()
            lstMove = [card, elem.position, elem.x, elem.y, str1, str2]
            mainWnd.lstMoves.append(lstMove)
            # проверка завершения игры:
            if self.isGameOver():
                mainWnd.gameOver()
            mainWnd.undoTheLastMoveAvailability()
        else:
            elem.setPixmap(elem.tempPixmap)
            elem.tempPixmap = None
            
class Dialog(QtGui.QDialog):
    def __init__(self, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.setMinimumWidth(300) # минимальная ширина окна диалога
        self.setFixedHeight(100) # запрет изменения высоты
        
        lay = QtGui.QGridLayout(self)
        label1 = QtGui.QLabel(parent.tr.translate('Name') + ':', self) # метка
        label1.setSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        lay.addWidget(label1, 0, 0)
        self.ln_edit1 = QtGui.QLineEdit('', self) # строковое поле ввода
        self.ln_edit1.setMaxLength(100)
        lay.addWidget(self.ln_edit1, 0, 1, 1, 2)
        
        boxlay = QtGui.QHBoxLayout()
        lay.addLayout(boxlay, 2, 2)
        for it in (['Ok', 1], ['Cancel', 0]): # кнопки
            button = QtGui.QPushButton(parent.tr.translate(it[0]), self)
            button.setSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
            self.connect(button, QtCore.SIGNAL('clicked()'), self.buttonPress)
            button.result = it[1] # monkey
            boxlay.addWidget(button)
            
        self.setWindowTitle(parent.tr.translate('Scores recording')) # заголовок окна
        
    def buttonPress(self):
        # Обработчик нажатия кнопок "Ok" и "Cancel".
        self.done(self.sender().result)
        
if __name__=="__main__":
    app = QtGui.QApplication(sys.argv)
    app.setStyle("Plastique")
    main = MainWindow()
    main.show()
    sys.exit(app.exec_())
